@@ -1,12 +1,14 @@
 # -*- coding: utf-8 -*-
-# PyExifTool <http://github.com/sylikc/pyexiftool>
-# Copyright 2012 Sven Marnach.
-# Copyright 2021 Kevin M (sylikc)
-
-# More contributors in the CHANGELOG for the pull requests
-
+#
 # This file is part of PyExifTool.
 #
+# PyExifTool <http://github.com/sylikc/pyexiftool>
+#
+# Copyright 2019-2022 Kevin M (sylikc)
+# Copyright 2012-2014 Sven Marnach
+#
+# Community contributors are listed in the CHANGELOG.md for the PRs
+#
 # PyExifTool is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
 # the Free Software Foundation, either version 3 of the licence, or
@@ -18,8 +20,9 @@
 #
 # See COPYING.GPL or COPYING.BSD for more details.
 
+
 """
-This submodule contains the core ``ExifTool`` class for PyExifTool
+This submodule contains the core ``ExifTool`` class of PyExifTool
 
 .. note::
 	:py:class:`exiftool.helper.ExifTool` class of this submodule is available in the ``exiftool`` namespace as :py:class:`exiftool.ExifTool`
@@ -45,8 +48,10 @@ import ctypes
 try:
 	# Optional UltraJSON library - ultra-fast JSON encoder/decoder, drop-in replacement
 	import ujson as json
+	JSONDecodeError = ValueError  # ujson doesn't throw json.JSONDecodeError, but ValueError when string is malformed
 except ImportError:
 	import json  # type: ignore   # comment related to https://github.com/python/mypy/issues/1153
+	from json import JSONDecodeError
 import warnings
 
 
@@ -61,7 +66,7 @@ from typing import Optional, List, Union
 # ---------- Library Package Imports ----------
 
 from . import constants
-from .exceptions import ExifToolVersionError, ExifToolRunning, ExifToolNotRunning, OutputEmpty, OutputNotJSON
+from .exceptions import ExifToolVersionError, ExifToolRunning, ExifToolNotRunning, ExifToolOutputEmptyError, ExifToolJSONInvalidError
 
 
 # ======================================================================================================================
@@ -141,33 +146,9 @@ class ExifTool(object):
 	.. _Exiftool print conversion FAQ: https://exiftool.org/faq.html#Q6
 
 
-	You can pass optional arguments to the constructor:
-
-	* ``executable`` (string): file name of the *exiftool* executable.
-	  The default value :py:attr:`exiftool.constants.DEFAULT_EXECUTABLE` will only work if the executable
-	  is in your ``PATH``.
-	  You can also specify the full path to the ``exiftool`` executable.
-	  See :py:attr:`executable` property for more details.
-	* ``common_args`` (list of strings): contains additional parameters for
-	  the stay-open instance of exiftool.  The default is ``-G`` and ``-n``.
-	  See :py:attr:`common_args` property for more details.
-	* ``win_shell``
-
-	  .. note::
-	    This parameter may be deprecated in the future
-
-	* ``config_file`` (string): file path to ``-config`` parameter when
-	  starting process.
-	  See :py:attr:`config_file` property for more details.
-	* ``encoding`` (string): encoding to be used when communicating with
-	  exiftool process.  By default, will use ``locale.getpreferredencoding()``
-	  See :py:attr:`encoding` property for more details
-	* ``logger`` (object):  Set a custom logger to log status and debug messages to.
-	  See :py:attr:`logger` for more details.
-
 	Some methods of this class are only available after calling
-	:py:meth:`run()`, which will actually launch the *exiftool* subprocess.  To
-	avoid leaving the subprocess running, make sure to call
+	:py:meth:`run()`, which will actually launch the *exiftool* subprocess.
+	To avoid leaving the subprocess running, make sure to call
 	:py:meth:`terminate()` method when finished using the instance.
 	This method will also be implicitly called when the instance is
 	garbage collected, but there are circumstance when this won't ever
@@ -184,6 +165,7 @@ class ExifTool(object):
 
 	.. warning::
 		Note that options and parameters are not checked.  There is no error handling or validation of options passed to *exiftool*.
+
 		Nonsensical options are mostly silently ignored by exiftool, so there's not
 		much that can be done in that regard.  You should avoid passing
 		non-existent files to any of the methods, since this will lead
@@ -201,13 +183,60 @@ class ExifTool(object):
 	  executable: Optional[str] = None,
 	  common_args: Optional[List[str]] = ["-G", "-n"],
 	  win_shell: bool = False,
-	  config_file: Optional[str] = None,
+	  config_file: Optional[Union[str, Path]] = None,
 	  encoding: Optional[str] = None,
 	  logger = None) -> None:
-		""" ``common_args`` defaults to *-G -n* as this is the most common use case.
+		"""
 
-		* -G separates the grouping
-		* -n (print conversion disabled) improves the speed and consistency of output, and is more machine-parsable
+		:param executable: Specify file name of the *exiftool* executable if it is in your ``PATH``.  Otherwise, specify the full path to the ``exiftool`` executable.
+
+			Passed directly into :py:attr:`executable` property.
+
+			.. note::
+				The default value :py:attr:`exiftool.constants.DEFAULT_EXECUTABLE` will only work if the executable is in your ``PATH``.
+
+		:type executable: str, or None to use default
+
+
+		:param common_args:
+			Pass in additional parameters for the stay-open instance of exiftool.
+
+			Defaults to ``['-G', '-n']`` as this is the most common use case.
+
+			* ``-G`` (groupName level 1 enabled) separates the output with *groupName:tag* to disambiguate same-named tags under different groups.
+			* ``-n`` (print conversion disabled) improves the speed and consistency of output, and is more machine-parsable
+
+			Passed directly into :py:attr:`common_args` property.
+
+
+			.. note::
+				Depending on your use case, there may be other useful grouping levels and options.  Search `Phil Harvey's exiftool documentation`_ for **groupNames** and **groupHeadings** to get more info.
+
+
+
+			.. _`Phil Harvey's exiftool documentation`: https://exiftool.org/exiftool_pod.html
+
+		:type common_args: list of str, or None.
+
+		:param bool win_shell: (Windows only) Minimizes the exiftool process.
+
+			.. note::
+				This parameter may be deprecated in the future
+
+		:param config_file:
+			File path to ``-config`` parameter when starting exiftool process.
+
+			Passed directly into :py:attr:`config_file` property.
+		:type config_file: str, Path, or None
+
+		:param encoding: Specify encoding to be used when communicating with
+			exiftool process.  By default, will use ``locale.getpreferredencoding()``
+
+			Passed directly into :py:attr:`encoding` property.
+
+		:param logger: Set a custom logger to log status and debug messages to.
+
+			Passed directly into :py:attr:`logger` property.
 		"""
 
 		# --- default settings / declare member variables ---
@@ -535,29 +564,39 @@ class ExifTool(object):
 
 	# ----------------------------------------------------------------------------------------------------------------------
 	@property
-	def last_stdout(self) -> Optional[str]:
+	def last_stdout(self) -> Optional[Union[str, bytes]]:
 		"""
 		``STDOUT`` for most recent result from execute()
 
+		.. note::
+			The return type can be either str or bytes.
+
+			If the most recent call to execute() ``raw_bytes=True``, then this will return ``bytes``.  Otherwise this will be ``str``.
+
 		.. note::
 			This property can be read at any time, and is not dependent on running state of ExifTool.
 
 			It is INTENTIONALLY *NOT* CLEARED on exiftool termination, to allow
-			for executing a command and terminating, but still have result available.
+			for executing a command and terminating, but still having the result available.
 		"""
 		return self._last_stdout
 
 	# ----------------------------------------------------------------------------------------------------------------------
 	@property
-	def last_stderr(self) -> Optional[str]:
+	def last_stderr(self) -> Optional[Union[str, bytes]]:
 		"""
 		``STDERR`` for most recent result from execute()
 
+		.. note::
+			The return type can be either ``str`` or ``bytes``.
+
+			If the most recent call to execute() ``raw_bytes=True``, then this will return ``bytes``.  Otherwise this will be ``str``.
+
 		.. note::
 			This property can be read at any time, and is not dependent on running state of ExifTool.
 
 			It is INTENTIONALLY *NOT* CLEARED on exiftool termination, to allow
-			for executing a command and terminating, but still have result available.
+			for executing a command and terminating, but still having the result available.
 		"""
 		return self._last_stderr
 
@@ -571,7 +610,7 @@ class ExifTool(object):
 			This property can be read at any time, and is not dependent on running state of ExifTool.
 
 			It is INTENTIONALLY *NOT* CLEARED on exiftool termination, to allow
-			for executing a command and terminating, but still have result available.
+			for executing a command and terminating, but still having the result available.
 		"""
 		return self._last_status
 
@@ -587,7 +626,7 @@ class ExifTool(object):
 		""" set a new user-created logging.Logger object
 			can be set at any time to start logging.
 
-			Set to None at any time to stop logging
+			Set to None at any time to stop logging.
 		"""
 		if new_logger is None:
 			self._logger = None
@@ -710,13 +749,19 @@ class ExifTool(object):
 
 
 		try:
+			# NOTE: the encoding= parameter was removed from the Popen() call to support
+			# using bytes in the actual communication with exiftool process.
+			# Due to the way the code is written, ExifTool only uses stdin.write which would need to be in bytes.
+			# The reading is _NOT_ using subprocess.communicate().  This class reads raw bytes using os.read()
+			# Therefore, by switching off the encoding= in Popen(), we can support both bytes and str at the
+			# same time.  (This change was to support https://github.com/sylikc/pyexiftool/issues/47)
+
 			# unify both platform calls into one subprocess.Popen call
 			self._process = subprocess.Popen(
 				proc_args,
 				stdin=subprocess.PIPE,
 				stdout=subprocess.PIPE,
 				stderr=subprocess.PIPE,
-				encoding=self._encoding,
 				**kwargs)
 		except FileNotFoundError as fnfe:
 			raise
@@ -794,13 +839,11 @@ class ExifTool(object):
 			try:
 				"""
 					On Windows, running this after __del__ freezes at communicate(), regardless of timeout
-						this is possibly because the file descriptors are no longer valid or were closed at __del__
-
-						test yourself with simple code that calls .run() and then end of script
+						see the bug filed above for details
 
 					On Linux, this runs as is, and the process terminates properly
 				"""
-				self._process.communicate(input="-stay_open\nFalse\n", timeout=timeout)  # TODO these are constants which should be elsewhere defined
+				self._process.communicate(input=b"-stay_open\nFalse\n", timeout=timeout)  # this is a constant sequence specified by PH's exiftool
 				self._process.kill()
 			except subprocess.TimeoutExpired:  # this is new in Python 3.3 (for python 2.x, use the PyPI subprocess32 module)
 				self._process.kill()
@@ -821,12 +864,12 @@ class ExifTool(object):
 	##################################################################################
 
 	# ----------------------------------------------------------------------------------------------------------------------
-	def execute(self, *params: str) -> str:
+	def execute(self, *params: Union[str, bytes], raw_bytes: bool = False) -> Union[str, bytes]:
 		"""Execute the given batch of parameters with *exiftool*.
 
 		This method accepts any number of parameters and sends them to
 		the attached ``exiftool`` subprocess.  The process must be
-		running, otherwise ``ExifToolNotRunning`` is raised.  The final
+		running, otherwise :py:exc:`exiftool.exceptions.ExifToolNotRunning` is raised.  The final
 		``-execute`` necessary to actually run the batch is appended
 		automatically; see the documentation of :py:meth:`run()` for
 		the common options.  The ``exiftool`` output is read up to the
@@ -834,9 +877,10 @@ class ExifTool(object):
 		based on the currently set :py:attr:`encoding`,
 		excluding the sentinel.
 
-		The parameters must be of type ``str``.  Use the :py:attr:`encoding` property to change the
-		encoding ``exiftool`` accepts.  For filenames, this should be the
-		system's filesystem encoding.
+		The parameters must be of type ``str`` or ``bytes``.
+		``str`` parameters are encoded to bytes automatically using the :py:attr:`encoding` property.
+		For filenames, this should be the system's filesystem encoding.
+		``bytes`` parameters are untouched and passed directly to ``exiftool``.
 
 		.. note::
 			This is the core method to interface with the ``exiftool`` subprocess.
@@ -848,10 +892,12 @@ class ExifTool(object):
 			Typically passed in via `Unpacking Argument Lists`_
 
 			.. note::
-				The parameters to this function must be type ``str``
+				The parameters to this function must be type ``str`` or ``bytes``.
 
 		:type params: one or more string parameters
 
+		:param raw_bytes: If True, returns bytes.  Default behavior returns a str
+
 
 		:return:
 			* STDOUT is returned by the method call, and is also set in :py:attr:`last_stdout`
@@ -860,6 +906,8 @@ class ExifTool(object):
 
 		:raises ExifToolNotRunning: If attempting to execute when not running (:py:attr:`running` == False)
 		:raises ExifToolVersionError: If unexpected text was returned from the command while parsing out the sentinels
+		:raises UnicodeDecodeError: If the :py:attr:`encoding` is not specified properly, it may be possible for ``.decode()`` method to raise this error
+		:raises TypeError: If ``params`` argument is not ``str`` or ``bytes``
 
 
 		.. _Unpacking Argument Lists: https://docs.python.org/3/tutorial/controlflow.html#unpacking-argument-lists
@@ -885,26 +933,69 @@ class ExifTool(object):
 		SEQ_ERR_STATUS_DELIM = "="  # this can be configured to be one or more chacters... the code below will accomodate for longer sequences: len() >= 1
 		seq_err_status = "${status}"  # a special sequence, ${status} returns EXIT STATUS as per exiftool documentation - only supported on exiftool v12.10+
 
+
+		# ---------- build the params list and encode the params to bytes, if necessary ----------
+		cmd_params: bytes = []
+
+		# this is necessary as the encoding parameter of Popen() is not specified.  We manually encode as per the .encoding() parameter
+		for p in params:
+			# we use isinstance() over type() not only for subclass, but
+			# according to https://switowski.com/blog/type-vs-isinstance
+			# isinstance() is 40% faster than type()
+			if isinstance(p, bytes):
+				# no conversion needed, pass in raw (caller has already encoded)
+				cmd_params.append(p)
+			elif isinstance(p, str):
+				# conversion needed, encode based on specified encoding
+				cmd_params.append(p.encode(self._encoding))
+			else:
+				# technically at this point we could support any object and call str()
+				# but leave this up to an extended class like ExifToolHelper()
+				raise TypeError(f"ERROR: Parameter was not bytes/str: {type(p)} => {p}")
+
 		# f-strings are faster than concatentation of multiple strings -- https://stackoverflow.com/questions/59180574/string-concatenation-with-vs-f-string
-		cmd_text = "\n".join(params + ("-echo4", f"{SEQ_ERR_STATUS_DELIM}{seq_err_status}{SEQ_ERR_STATUS_DELIM}{seq_err_post}", seq_execute))
+		cmd_params.extend(
+			(b"-echo4",
+			f"{SEQ_ERR_STATUS_DELIM}{seq_err_status}{SEQ_ERR_STATUS_DELIM}{seq_err_post}".encode(self._encoding),
+			seq_execute.encode(self._encoding))
+		)
+
+		cmd_bytes = b"\n".join(cmd_params)
 
 
 		# ---------- write to the pipe connected with exiftool process ----------
 
-		self._process.stdin.write(cmd_text)
+		self._process.stdin.write(cmd_bytes)
 		self._process.stdin.flush()
 
-		if self._logger: self._logger.info("Method 'execute': Command sent = {}".format(cmd_text.split('\n')[:-1]))
+		if self._logger: self._logger.info("Method 'execute': Command sent = {}".format(cmd_params[:-1]))  # logs without the -execute (it would confuse people to include that)
 
 
 		# ---------- read output from exiftool process until special sequences reached ----------
 
+		# NOTE:
+		#
+		# while subprocess recommends: "Use communicate() rather than .stdin.write, .stdout.read or .stderr.read to avoid deadlocks due to any of the other OS pipe buffers filling up and blocking the child process."
+		#
+		# this raw reading is used instead of Popen.communicate() due to the note:
+		# https://docs.python.org/3/library/subprocess.html#subprocess.Popen.communicate
+		#
+		# "The data read is buffered in memory, so do not use this method if the data size is large or unlimited."
+		#
+		# The data that comes back from exiftool falls into this, and so unbuffered reads are done with os.read()
+
 		fdout = self._process.stdout.fileno()
-		raw_stdout = _read_fd_endswith(fdout, seq_ready.encode(self._encoding), self._block_size).decode(self._encoding)
+		raw_stdout = _read_fd_endswith(fdout, seq_ready.encode(self._encoding), self._block_size)
 
 		# when it's ready, we can safely read all of stderr out, as the command is already done
 		fderr = self._process.stderr.fileno()
-		raw_stderr = _read_fd_endswith(fderr, seq_err_post.encode(self._encoding), self._block_size).decode(self._encoding)
+		raw_stderr = _read_fd_endswith(fderr, seq_err_post.encode(self._encoding), self._block_size)
+
+
+		if not raw_bytes:
+			# decode if not returning bytes
+			raw_stdout = raw_stdout.decode(self._encoding)
+			raw_stderr = raw_stderr.decode(self._encoding)
 
 
 		# ---------- parse output ----------
@@ -913,16 +1004,21 @@ class ExifTool(object):
 		cmd_stdout = raw_stdout.strip()[:-len(seq_ready)]
 		cmd_stderr = raw_stderr.strip()[:-len(seq_err_post)]  # save it in case the error below happens and output can be checked easily
 
+
+		# if raw_bytes is True, the check has to become bytes rather than str
+		err_status_delim = SEQ_ERR_STATUS_DELIM if not raw_bytes else SEQ_ERR_STATUS_DELIM.encode(self._encoding)
+
+
 		# sanity check the status code from the stderr output
-		delim_len = len(SEQ_ERR_STATUS_DELIM)
-		if cmd_stderr[-delim_len:] != SEQ_ERR_STATUS_DELIM:
-			# exiftool is expected to dump out the status code within the delims... if it doesn't, the class is broken
-			raise ExifToolVersionError(f"Exiftool expected to return status on stderr, but got unexpected character: {cmd_stderr[-delim_len:]} != {SEQ_ERR_STATUS_DELIM}")
+		delim_len = len(err_status_delim)
+		if cmd_stderr[-delim_len:] != err_status_delim:
+			# exiftool is expected to dump out the status code within the delims... if it doesn't, the class doesn't match expected exiftool output for current version
+			raise ExifToolVersionError(f"Exiftool expected to return status on stderr, but got unexpected character: {cmd_stderr[-delim_len:]} != {err_status_delim}")
 
 		# look for the previous delim (we could use regex here to do all this in one step, but it's probably overkill, and could slow down the code significantly)
-		# the other simplification that can be done is that, Exiftool is expected to only return 0, 1, or 2 as per documentation
+		# the other simplification that can be done is that, as of this writing: Exiftool is expected to only return 0, 1, or 2 as per documentation
 		# you could just lop the last 3 characters off... but if the return status changes in the future, then this code would break
-		err_delim_1 = cmd_stderr.rfind(SEQ_ERR_STATUS_DELIM, 0, -delim_len)
+		err_delim_1 = cmd_stderr.rfind(err_status_delim, 0, -delim_len)
 		cmd_status = cmd_stderr[err_delim_1 + delim_len : -delim_len]
 
 
@@ -937,9 +1033,10 @@ class ExifTool(object):
 
 
 		if self._logger:
-			self._logger.debug(f"Method 'execute': Reply stdout = \"{self._last_stdout}\"")
-			self._logger.debug(f"Method 'execute': Reply stderr = \"{self._last_stderr}\"")
-			self._logger.debug(f"Method 'execute': Reply status = {self._last_status}")
+			self._logger.debug(f"{self.__class__.__name__}.execute: IN  params = {params}")
+			self._logger.debug(f"{self.__class__.__name__}.execute: OUT stdout = \"{self._last_stdout}\"")
+			self._logger.debug(f"{self.__class__.__name__}.execute: OUT stderr = \"{self._last_stderr}\"")
+			self._logger.debug(f"{self.__class__.__name__}.execute: OUT status = {self._last_status}")
 
 
 		# the standard return: just stdout
@@ -967,6 +1064,13 @@ class ExifTool(object):
 
 			You can adjust the output structure with various *exiftool* options.
 
+		.. warning::
+			This method does not verify the exit status code returned by *exiftool*.  That is left up to the caller.
+
+			This will mimic exiftool's default method of operation "continue on error" and "best attempt" to complete commands given.
+
+			If you want automated error checking, use :py:class:`exiftool.ExifToolHelper`
+
 		:param params: One or more parameters to send to the ``exiftool`` subprocess.
 
 			Typically passed in via `Unpacking Argument Lists`_
@@ -977,15 +1081,23 @@ class ExifTool(object):
 		:type params: one or more string parameters
 
 		:return: Valid JSON parsed into a Python list of dicts
-		:raises OutputEmpty: If *exiftool* did not return any STDOUT
+		:raises ExifToolOutputEmptyError: If *exiftool* did not return any STDOUT
 
 			.. note::
-				This is not necessarily an error, setting tags can cause this behavior.  Use :py:meth:`execute()` to set tags.
+				This is not necessarily an *exiftool* error, but rather a programmer error.
 
-		:raises OutputNotJSON: If *exiftool* returned STDOUT which is invalid JSON.
+				For example, setting tags can cause this behavior.
+
+				If you are executing a command and expect no output, use :py:meth:`execute()` instead.
+
+		:raises ExifToolJSONInvalidError: If *exiftool* returned STDOUT which is invalid JSON.
 
 			.. note::
-				This is not necessarily an error, ``-w`` can cause this behavior.  Use :py:meth:`execute()` if using the ``-w`` flag.
+				This is not necessarily an *exiftool* error, but rather a programmer error.
+
+				For example, ``-w`` can cause this behavior.
+
+				If you are executing a command and expect non-JSON output, use :py:meth:`execute()` instead.
 
 
 		.. _Unpacking Argument Lists: https://docs.python.org/3/tutorial/controlflow.html#unpacking-argument-lists
@@ -993,12 +1105,12 @@ class ExifTool(object):
 
 		result = self.execute("-j", *params)  # stdout
 
-		# TODO check status code or have caller do it?
-		"""
-		status_code = self._last_status  # status code
-		if status_code != 0:
-			warnings.warn(f"ExifTool returned a non-zero status code: {status_code}", UserWarning)
-		"""
+		# NOTE: I have decided NOT to check status code
+		# There are quite a few use cases where it's desirable to have continue-on-error behavior,
+		# as that is exiftool's default mode of operation.  exiftool normally just does what it can
+		# and tells you that it did all this and that, but some files didn't process.  In this case
+		# exit code is non-zero, but exiftool did SOMETHING.  I leave it up to the caller to figure
+		# out what was done or not done.
 
 
 		if len(result) == 0:
@@ -1012,12 +1124,13 @@ class ExifTool(object):
 
 			# Returning [] could be ambiguous if Exiftool changes the returned JSON structure in the future
 			# Returning None was preferred, because it's the safest as it clearly indicates that nothing came back from execute(), but it means execute_json() doesn't always return JSON
-			raise OutputEmpty("ExifTool did not return any stdout")
+			# Raising an error is the current solution, as that clearly indicates that you used execute_json() expecting output, but got nothing
+			raise ExifToolOutputEmptyError(self._last_status, self._last_stdout, self._last_stderr, params)
 
 
 		try:
 			parsed = json.loads(result)
-		except json.JSONDecodeError as e:
+		except JSONDecodeError as e:
 			# if `-w` flag is specified in common_args or params, stdout will not be JSON parseable
 			#
 			# which will return something like:
@@ -1027,7 +1140,7 @@ class ExifTool(object):
 			# the user is expected to know this ahead of time, and if -w exists in common_args or as a param, this error will be thrown
 
 			# explicit chaining https://www.python.org/dev/peps/pep-3134/
-			raise OutputNotJSON() from e
+			raise ExifToolJSONInvalidError(self._last_status, self._last_stdout, self._last_stderr, params) from e
 
 		return parsed
 
